----------------------------------------------------------------------------
-- LuaJIT LoongArch disassembler module.
--
-- Copyright (C) 2005-2021 Mike Pall. All rights reserved.
-- Released under the MIT/X license. See Copyright Notice in luajit.h
----------------------------------------------------------------------------
-- This is a helper module used by the LuaJIT machine code dumper module.
--
-- It disassembles most LoongArch instructions.
-- NYI: SIMD instructions.
------------------------------------------------------------------------------

local type = type
local byte, format = string.byte, string.format
local match, gmatch = string.match, string.gmatch
local concat = table.concat
local bit = require("bit")
local band, bor, tohex = bit.band, bit.bor, bit.tohex
local lshift, rshift, arshift = bit.lshift, bit.rshift, bit.arshift

------------------------------------------------------------------------------
-- Opcode maps
------------------------------------------------------------------------------

local map_18_0 = {      -- 18-20:0, 10-17
  shift = 10, mask = 255,
  [4] = "clo.wDJ",      "clz.wDJ",      "cto.wDJ",      "ctz.wDJ",
  "clo.dDJ",            "clz.dDJ",      "cto.dDJ",      "ctz.dDJ",
  "revb.2hDJ",          "revb.4hDJ",    "revb.2wDJ",    "revb.dDJ",
  "revh.2wDJ",          "revh.dDJ",     "bitrev.4bDJ",  "bitrev.8bDJ",
  "bitrev.wDJ",         "bitrev.dDJ",   "ext.w.hDJ",    "ext.w.bDJ",
}

local map_18_4 = {	-- 18-20:4, 15-17
  shift = 15, mask = 7,
  [0] = "add.wDJK",	"add.dDJK",	"sub.wDJK",	"sub.dDJK",
  "sltDJK",		"sltuDJK",	"maskeqzDJK",	"masknezDJK",
}

local map_18_5 = {	-- 18-20:5, 15-17
  shift = 15, mask = 7,
  [0] = "norDJK",	"andDJK",	"orDJK",	"xorDJK",
  "ornDJK",		"andnDJK",	"sll.wDJK",	"srl.wDJK",
}

local map_18_6 = {	-- 18-20:6, 15-17
  shift = 15, mask = 7,
  [0] = "sra.wDJK",	"sll.dDJK",	"srl.dDJK",	"sra.dDJK",
  false,		false,		"rotr.wDJK",	"rotr.dDJK",
}

local map_18_7 = {	-- 18-20:7, 15-17
  shift = 15, mask = 7,
  [0] = "mul.wDJK",	"mulh.wDJK",	"mulh.wuDJK",	"mul.dDJK",
  "mulh.dDJK",		"mulh.duDJK",	"mulw.d.wDJK",	"mulw.d.wuDJK",
}

local map_farith2 = {
  shift = 10, mask = 31,
  [0] = "fabs.hFG",     "fabs.sFG",     "fabs.dFG",     false,
  "fneg.hFG",           "fneg.sFG",     "fneg.dFG",     false,
  "flogb.hFG",          "flogb.sFG",    "flogb.dFG",    false,
  "fclass.hFG",         "fclass.sFG",   "fclass.dFG",   false,
  "fsqrt.hFG",          "fsqrt.sFG",    "fsqrt.dFG",    false,
  "frecip.hFG",         "frecip.sFG",   "frecip.dFG",   false,
  "frsqrt.hFG",         "frsqrt.sFG",   "frsqrt.dFG",   false,
  "frecipe.hFG",        "frecipe.sFG",  "frecipe.dFG",  false,
  "frsqrte.hFG",        "frsqrte.sFG",  "frsqrte.dFG",
}

local map_fmov = {
  shift = 10, mask = 31,
  [4] = "fmov.hFG", "fmov.sFG",     "fmov.dFG",     false,
  "movgr2fr.hFJ",       "movgr2fr.wFJ", "movgr2fr.dFJ", "movgr2frh.wFJ",
  "movfr2gr.hDG",       "movfr2gr.sDG", "movfr2gr.dDG", "movfrh2gr.sDG",
  "movgr2fcsrSJ",       false,          "movfcsr2grDR", false,
  { shift = 3, mask = 3, [0] = "movfr2cfEG", },
  { shift = 8, mask = 3, [0] = "movcf2frFA", },
  { shift = 3, mask = 3, [0] = "movgr2cfEJ", },
  { shift = 8, mask = 3, [0] = "movcf2grDA", },
  "fcvt.ld.dFG",        "fcvt.ud.dFG",
}

local map_fconvert = { -- 15-20: 110010
  shift = 10, mask = 31,
  [6] = "fcvt.s.dFG",	[9] = "fcvt.d.sFG",
}

local map_fconvert1 = { -- 15-20: 110100
  shift = 10, mask = 31,
  [1] = "ftintrm.w.sFG", "ftintrm.w.dFG",
  [9] = "ftintrm.l.sFG", "ftintrm.l.dFG",
  [17] = "ftintrp.w.sFG", "ftintrp.w.dFG",
  [25] = "ftintrp.l.sFG", "ftintrp.l.dFG",
}

local map_fconvert2 = { -- 15-20: 110101
  shift = 10, mask = 31,
  [1] = "ftintrz.w.sFG", "ftintrz.w.dFG",
  [9] = "ftintrz.l.sFG", "ftintrz.l.dFG",
  [17] = "ftintrne.w.sFG", "ftintrne.w.dFG",
  [25] = "ftintrne.l.sFG", "ftintrne.l.dFG",
}

local map_fconvert3 = { -- 15-20: 110110
  shift = 10, mask = 31,
  [1] = "ftint.w.sFG", "ftint.w.dFG",
  [9] = "ftint.l.sFG", "ftint.l.dFG",
}

local map_fconvert4 = { -- 15-20: 111010
  shift = 10, mask = 31,
  [4] = "ffint.s.wFG", false, "ffint.s.lFG", false,
  "ffint.d.wFG", false, "ffint.d.lFG"
}

local map_fconvert5 = { -- 15-20: 111100
  shift = 10, mask = 31,
  [17] = "frint.sFG", "frint.dFG",
}

local map_farith = {	-- 22-25:4, 15-21
  shift = 15, mask = 127,
  [0] = "fadd.hFGH",	"fadd.sFGH",	"fadd.dFGH",	false,
  "fsub.hFGH",		"fsub.sFGH",	"fsub.dFGH",	false,
  "fmul.hFGH",		"fmul.sFGH",	"fmul.dFGH",	false,
  "fdiv.hFGH",		"fdiv.sFGH",	"fdiv.dFGH",	false,
  "fmax.hFGH",		"fmax.sFGH",	"fmax.dFGH",	false,
  "fmin.hFGH",		"fmin.sFGH",	"fmin.dFGH",	false,
  "fmaxa.hFGH",		"fmaxa.sFGH",	"fmaxa.dFGH",	false,
  "fmina.hFGH",		"fmina.sFGH",	"fmina.dFGH",	false,
  "fscaleb.hFGH",	"fscaleb.sFGH",	"fscaleb.dFGH",	false,
  "fcopysign.hFGH",	"fcopysign.sFGH",	"fcopysign.dFGH",
  [40] = map_farith2, [41] = map_fmov,
  [50] = map_fconvert, [52] = map_fconvert1,
  [53] = map_fconvert2, [54] = map_fconvert3,
  [58] = map_fconvert4, [60] = map_fconvert5,
}

local map_21_0 = {	--21st:0, 18-20
  shift = 18, mask = 7,
  [0] = map_18_0,
  [1] = { shift = 17, mask = 1, [0] = "alsl.wDJKQ", "alsl.wuDJKQ", },
  [2] = {shift = 17, mask = 1, [0] = "bytepick.wDJKQ", },
  [3] = "bytepick.dDJKB",
  [4] = map_18_4, map_18_5, map_18_6, map_18_7,
}

local map_21_1 = {      --21st:1, 22nd:0, 15-20
  shift = 21, mask = 1,
  [1] = {
    shift = 18, mask = 7,
    [0] = {
      shift = 15, mask = 7,
      [0] = "div.wDJK",   "mod.wDJK",     "div.wuDJK",    "mod.wuDJK",
      "div.dDJK",         "mod.dDJK",     "div.duDJK",    "mod.duDJK",
    },
    [1] = {
      shift = 18, mask = 7,
      "crc.w.b.wDJK",     "crc.w.h.wDJK", "crc.w.w.wDJK", "crc.w.d.wDJK",
      "crcc.w.b.wDJK",    "crcc.w.h.wDJK",        "crcc.w.w.wDJK",        "crcc.w.d.wDJK",
    },
    [2] = {
      shift = 15, mask = 7,
      [4] = breakC, [6] = syscallC,
    },
    [3] = { shift = 17, mask = 1, [0] = "alsl.dDJKQ", },
  },
}

local map_22_0 = {
  shift = 21, mask = 1,
  [0] = map_21_0,
  [1] = map_21_1,
}

local map_shift = {	-- 22nd:1, 21st:0
  shift = 16, mask = 31,
  [0] = { shift = 15, mask = 1, [1] = "slli.wDJU", },
  [1] = "slli.dDJV",
  [4] = { shift = 15, mask = 1, [1] = "srli.wDJU", },
  [5] = "srli.dDJV",
  [8] = { shift = 15, mask = 1, [1] = "srai.wDJU", },
  [9] = "srai.dDJV",
  [12] = { shift = 15, mask = 1, [1] = "rotri.wDJU", },
  [13] = "rotri.dDJV",
}

local map_22_1 = {        -- 22nd:1
  shift = 21, mask = 1,
  [0] = map_shift,
  [1] = { shift = 15, mask = 1, [0] = "bstrins.wDJMU", "bstrpick.wDJMU", },
}

local map_26_0 = {
  shift = 22, mask = 15,
  [0] = map_22_0,
  [1] = map_22_1,
  [2] = "bstrins.dDJNV", "bstrpick.dDJNV",
  [4] = map_farith,
  [8] = "sltiDJX",      "sltuiDJX",     "addi.wDJX",    "addi.dDJX",
  "lu52i.dDJX",         "andiDJT",      "oriDJT",       "xoriDJT",
}

local map_long_i_5 = { -- Long immediate fixed-point arithmetic.
  shift = 25, mask = 1,
  [0] = "lu12i.wDZ",	"lu32i.dDZ",
}

local map_long_i_6 = {
  shift = 25, mask = 1,
  [0] = "pcaddiDZ",	"pcalau12iDZ",
}

local map_long_i_7 = {
  shift = 25, mask = 1,
  [0] = "pcaddu12iDZ",	"pcaddu18iDZ",
}

local map_ldst0_14 = {
  shift = 15, mask = 2047,
  [0] = "ldx.bDJK", [8] = "ldx.hDJK", [16] = "ldx.wDJK",
  [24] = "ldx.dDJK", [32] = "stx.bDJK", [40] = "stx.hDJK",
  [48] = "stx.wDJK", [56] = "stx.dDJK", [64] = "ldx.buDJK",
  [72] = "ldx.huDJK", [80] = "ldx.wuDJK", [96] = "fldx.sFJK",
  [104] = "fldx.dFJK", [112] = "fstx.sFJK", [120] = "fstx.dFJK",
  [232] = "fldgt.sFJK", "fldgt.dFJK", "fldle.sFJK", "fldle.dFJK",
  "fstgt.sFJK", "fstgt.dFJK", "fstle.sFJK", "fstle.dFJK",
  "ldgt.bDJK", "ldgt.hDJK", "ldgt.wDJK", "ldgt.dDJK",
  "ldle.bDJK", "ldle.hDJK", "ldle.wDJK", "ldle.dDJK",
  "stgt.bDJK", "stgt.hDJK", "stgt.wDJK", "stgt.dDJK",
  "stle.bDJK", "stle.hDJK", "stle.wDJK", "stle.dDJK",
}

local map_ldst1_8 = {
  shift = 24, mask = 3,
  [0] = "ll.wDJW",	"sc.wDJW",	"ll.dDJW",	"sc.dDJW",
}

local map_ldst1_9 = {
  shift = 24, mask = 3,
  [0] = "ldptr.wDJW",	"stptr.wDJW",	"ldptr.dDJW",	"stptr.dDJW",
}

local map_ldst1_10 = {
  shift = 22, mask = 15,
  [0] = "ld.bDJX",	"ld.hDJX",	"ld.wDo",	"ld.dDo",
  "st.bDo",		"st.hDo",	"st.wDo",	"st.dDo",
  "ld.buDo",		"ld.huDo",	"ld.wuDJX",	false,
  "fld.sFo",		"fst.sFo",	"fld.dFo",	"fst.dFo",
}

local map_fcmp0 = {
  shift = 15, mask = 31,
  [0] = "fcmp.caf.sEGH",  "fcmp.saf.sEGH",  "fcmp.clt.sEGH",  "fcmp.slt.sEGH",
  "fcmp.ceq.sEGH",        "fcmp.seq.sEGH",  "fcmp.cle.sEGH",  "fcmp.sle.sEGH",
  "fcmp.cun.sEGH",        "fcmp.sun.sEGH",  "fcmp.cult.sEGH", "fcmp.sult.sEGH",
  "fcmp.cueq.sEGH",       "fcmp.sueq.sEGH", "fcmp.cule.sEGH", "fcmp.sule.sEGH",
  "fcmp.cne.sEGH",        "fcmp.sne.sEGH",  false,            false,
  "fcmp.cor.sEGH",        "fcmp.sor.sEGH",  false,            false,
  "fcmp.cune.sEGH",       "fcmp.sune.sEGH",
}

local map_fcmp1 = {
  shift = 15, mask = 31,
  [0] = "fcmp.caf.dEGH",  "fcmp.saf.dEGH",  "fcmp.clt.dEGH",  "fcmp.slt.dEGH",
  "fcmp.ceq.dEGH",        "fcmp.seq.dEGH",  "fcmp.cle.dEGH",  "fcmp.sle.dEGH",
  "fcmp.cun.dEGH",        "fcmp.sun.dEGH",  "fcmp.cult.dEGH", "fcmp.sult.dEGH",
  "fcmp.cueq.dEGH",       "fcmp.sueq.dEGH", "fcmp.cule.dEGH", "fcmp.sule.dEGH",
  "fcmp.cne.dEGH",        "fcmp.sne.dEGH",  false,            false,
  "fcmp.cor.dEGH",        "fcmp.sor.dEGH",  false,            false,
  "fcmp.cune.dEGH",       "fcmp.sune.dEGH",
}

local map_fcmp = {
  shift = 20, mask = 63,
  [1] = { shift = 3, mask = 3, [0] = map_fcmp0, },
  [2] = { shift = 3, mask = 3, [0] = map_fcmp1, },
  [16] = { shift = 18, mask = 3, [0] = "fselFGHI", },
}

local map_init = {
  shift = 26, mask = 63,
  [0] = map_26_0,
  [3] = map_fcmp,
  [4] = "addu16i.dDJY",
  [5] = map_long_i_5,
  [6] = map_long_i_6,
  [7] = map_long_i_7,
  [8] = map_ldst1_8, [9] = map_ldst1_9, [10] = map_ldst1_10,
  [14] = map_ldst0_14,
  [16] = "beqzJL", "bnezJL",
  [18] = { shift = 8, mask = 3, [0] = "bceqzAL", "bcnezAL", },
  [19] = "jirlDJa", "bP", "blP", "beqJDO", "bneJDO", "bltJDO",
  "bgeJDO", "bltuJDO", "bgeuJDO",
}

------------------------------------------------------------------------------

local map_gpr = {
  [0] = "r0", "ra", "r2", "sp", "r4", "r5", "r6", "r7",
  "r8", "r9", "r10", "r11", "r12", "r13", "r14", "r15",
  "r16", "r17", "r18", "r19", "r20", "r21", "r22", "r23",
  "r24", "r25", "r26", "r27", "r28", "r29", "r30", "r31",
}

------------------------------------------------------------------------------

-- Output a nicely formatted line with an opcode and operands.
local function putop(ctx, text, operands)
  local pos = ctx.pos
  local extra = ""
  if ctx.rel then
    local sym = ctx.symtab[ctx.rel]
    if sym then extra = "\t->"..sym end
  end
  if ctx.hexdump > 0 then
    ctx.out(format("%08x  %s  %-7s %s%s\n",
	    ctx.addr+pos, tohex(ctx.op), text, concat(operands, ", "), extra))
  else
    ctx.out(format("%08x  %-7s %s%s\n",
	    ctx.addr+pos, text, concat(operands, ", "), extra))
  end
  ctx.pos = pos + 4
end

-- Fallback for unknown opcodes.
local function unknown(ctx)
  return putop(ctx, ".long", { "0x"..tohex(ctx.op) })
end

local function get_le(ctx)
  local pos = ctx.pos
  local b0, b1, b2, b3 = byte(ctx.code, pos+1, pos+4)
  return bor(lshift(b3, 24), lshift(b2, 16), lshift(b1, 8), b0)
end

-- Disassemble a single instruction.
local function disass_ins(ctx)
  local op = ctx:get()
  local operands = {}
  local last = nil
  ctx.op = op
  ctx.rel = nil

  local opat = ctx.map_pri[rshift(op, 26)]
  while type(opat) ~= "string" do
    if not opat then return unknown(ctx) end
    opat = opat[band(rshift(op, opat.shift), opat.mask)]
  end
  local name, pat = match(opat, "^([a-z0-9_.]*)(.*)")
  local altname, pat2 = match(pat, "|([a-z0-9_.|]*)(.*)")
  if altname then pat = pat2 end

  for p in gmatch(pat, ".") do
    local x = nil
    if p == "D" then
      x = map_gpr[band(rshift(op, 0), 31)]
    elseif p == "J" then
      x = map_gpr[band(rshift(op, 5), 31)]
    elseif p == "K" then
      x = map_gpr[band(rshift(op, 10), 31)]
    elseif p == "F" then
      x = "f"..band(rshift(op, 0), 31)
    elseif p == "G" then
      x = "f"..band(rshift(op, 5), 31)
    elseif p == "H" then
      x = "f"..band(rshift(op, 10), 31)
    elseif p == "S" then
      x = "fcsr"..band(rshift(op, 0), 31)
    elseif p == "R" then
      x = "fcsr"..band(rshift(op, 5), 31)
    elseif p == "E" then
      x = "fcc"..band(rshift(op, 0), 7)
    elseif p == "A" then
      x = "fcc"..band(rshift(op, 5), 7)
    elseif p == "I" then
      x = "fcc"..band(rshift(op, 15), 7)
    elseif p == "Q" then	--TODO sa2
      x = band(rshift(op, 15), 3)
    elseif p == "B" then	--TODO sa3
      x = band(rshift(op, 15), 7)
    elseif p == "M" then	--TODO msbw
      x = band(rshift(op, 16), 31)
    elseif p == "N" then	--TODO msbd
      x = band(rshift(op, 16), 63)
    elseif p == "U" then	-- ui5
      x = band(rshift(op, 10), 31)
    elseif p == "V" then	-- ui6
      x = band(rshift(op, 10), 63)
    elseif p == "T" then	-- ui12
      x = band(rshift(op, 10), 4095)
    elseif p == "W" then	-- si14
      x = band(rshift(op, 10), 16383)
    elseif p == "X" then	-- si12
      x = band(rshift(op, 10), 4095)
    elseif p == "o" then
      local disp = band((rshift(op, 10)), 0xfff)
      operands[#operands] = format("%s, %d", last, disp)
    elseif p == "Y" then	-- si16
      x = band(rshift(op, 10), 65535)
    elseif p == "Z" then	-- si20
      x = band(rshift(op, 10), 1048575)
    elseif p == "C" then	-- code
      x = band(rshift(op, 0), 32767)
    elseif p == "O" then	-- offs[15:0]
      x = band(rshift(op, 10), 65535)
    elseif p == "L" then	-- offs[15:0] + offs[20:16]
      x = lshift(band(op, 31), 16) + band(rshift(op, 10), 65535)
    elseif p == "P" then	-- offs[15:0] + offs[25:16]
      x = lshift(band(op, 1023), 16) + band(rshift(op, 10), 65535)
    elseif p == "a" then
      x = band(rshift(op, 10), 65535)
    else
      assert(false)
    end
    if x then operands[#operands+1] = x; last = x end
  end

  return putop(ctx, name, operands)
end

------------------------------------------------------------------------------

-- Disassemble a block of code.
local function disass_block(ctx, ofs, len)
  if not ofs then ofs = 0 end
  local stop = len and ofs+len or #ctx.code
  stop = stop - stop % 4
  ctx.pos = ofs - ofs % 4
  ctx.rel = nil
  while ctx.pos < stop do disass_ins(ctx) end
end

-- Extended API: create a disassembler context. Then call ctx:disass(ofs, len).
local function create(code, addr, out)
  local ctx = {}
  ctx.code = code
  ctx.addr = addr or 0
  ctx.out = out or io.write
  ctx.symtab = {}
  ctx.disass = disass_block
  ctx.hexdump = 8
  ctx.get = get_le
  ctx.map_pri = map_init
  return ctx
end

-- Simple API: disassemble code (a string) at address and output via out.
local function disass(code, addr, out)
  create(code, addr, out):disass()
end

-- Return register name for RID.
local function regname(r)
  if r < 32 then return map_gpr[r] end
  return "f"..(r-32)
end

-- Public module functions.
return {
  create = create,
  disass = disass,
  regname = regname
}

